#!/usr/bin/perl
# Copyright (c) 2004 Gavin Brown. This program is free software, you can use it
# and/or modify it under the same terms as Perl itself.
#
# $Id: gresolver.pl,v 1.3 2004/10/17 12:43:22 jodrell Exp $
use Gtk2 -init;
use Gtk2::GladeXML;
use Gtk2::Helper;
use File::Temp qw(tmpnam);
use POSIX qw(setlocale);
use Locale::gettext;
use strict;

chomp(my $dig = `which dig 2>/dev/null`);

### we need an executable dig, pop up an error dialog if one can't be found:
if (!-x $dig) {
	my $dialog = Gtk2::MessageDialog->new(undef, 'modal', 'error', 'ok', gettext("Can't find the 'dig' program!"));
	$dialog->signal_connect('response', sub { exit 1 });
	$dialog->signal_connect('delete_event', sub { exit 1 });
	$dialog->run;
}

### static variables:
my $NAME	= 'GResolver';
my $SNAME	= lc($NAME);
my $VERSION	= '0.0.1';
my $RCFILE	= sprintf('%s/.%src', $ENV{HOME}, $SNAME);
my $OPTIONS	= load_options();
my @SERVER_HIST	= split(/\|/, $OPTIONS->{server_hist});
my @QUERY_HIST	= split(/\|/, $OPTIONS->{query_hist});
my $PREFIX	= '@PREFIX@';
my $GLADE_FILE	= (-d $PREFIX ? "$PREFIX/share/$SNAME"	: $ENV{PWD})."/$SNAME.glade";
my $ICON_FILE	= (-d $PREFIX ? "$PREFIX/share/pixmaps"	: $ENV{PWD})."/$SNAME.png";
my $LOCALE_DIR	= (-d $PREFIX ? "$PREFIX/share/locale"	: $ENV{PWD}.'/locale');
my $LOCALE	= (defined($ENV{LC_MESSAGES}) ? $ENV{LC_MESSAGES} : $ENV{LANG});

setlocale(LC_ALL, $LOCALE);
bindtextdomain($SNAME, $LOCALE_DIR);
textdomain($SNAME);

my $MARKUP	= <<"END";
<span weight="bold" size="large">$NAME</span>
<span size="small">v$VERSION</span>

Copyright (c) 2004 Gavin Brown. This program is free software, you can use it and/or modify it under the same terms as Perl itself.
END
my $normal	= Gtk2::Gdk::Cursor->new('left_ptr');
my $busy	= Gtk2::Gdk::Cursor->new('watch');

### build the UI:
my $glade = Gtk2::GladeXML->new($GLADE_FILE);
$glade->signal_autoconnect_from_package('main');

### the application icon:
my $ICON = Gtk2::Gdk::Pixbuf->new_from_file($ICON_FILE);

### set up the about dialog:
$glade->get_widget('about_icon')->set_from_pixbuf($ICON);
$glade->get_widget('about_label')->set_markup($MARKUP);
$glade->get_widget('about_dialog')->set_icon($ICON);

### set up the output treeview:
my @cols = qw(Host TTL Class Type Answer);
my $store = Gtk2::TreeStore->new(qw/Glib::String Glib::String Glib::String Glib::String Glib::String/);
$glade->get_widget('output')->set_model($store);
for (0..4) {
	$glade->get_widget('output')->insert_column_with_attributes(
		$_,
		gettext($cols[$_]),
		Gtk2::CellRendererText->new,
		text => $_,
	);
}
foreach my $col ($glade->get_widget('output')->get_columns) {
	$col->set_resizable(1);
}

### setup the raw output textview:
$glade->get_widget('raw_response')->get_buffer->create_tag(
	'monospace',
	font		=> $OPTIONS->{font},
	wrap_mode	=> 'none',
);


$glade->get_widget('query_combo')->disable_activate;
$glade->get_widget('server_combo')->disable_activate;

### apply user settings:
$glade->get_widget('main_window')->set_position('center');
$glade->get_widget('main_window')->set_icon($ICON);

$glade->get_widget('query_combo')->set_popdown_strings('', @QUERY_HIST);
$glade->get_widget('server_combo')->set_popdown_strings('', @SERVER_HIST);

$glade->get_widget('recursive_checkbutton')->set_active($OPTIONS->{recursive_on} == 1);
$glade->get_widget('additional_checkbutton')->set_active($OPTIONS->{additional_on} == 1);
$glade->get_widget('authoritative_checkbutton')->set_active($OPTIONS->{authoritative_on} == 1);
$glade->get_widget('trace_checkbutton')->set_active($OPTIONS->{trace_on} == 1);
$glade->get_widget('options_expander')->set_expanded($OPTIONS->{options_visible} == 1);
$glade->get_widget('notebook')->set_current_page($OPTIONS->{current_page});
$glade->get_widget('type_combo')->set_active($OPTIONS->{last_type});

$glade->get_widget('query_button')->set_sensitive(undef);

$glade->get_widget('main_window')->show_all;

Gtk2->main;

exit 0;

### gather user settings and save config:
sub close_program {
	$OPTIONS->{recursive_on}	= ($glade->get_widget('recursive_checkbutton')->get_active	? 1 : 0);
	$OPTIONS->{additional_on}	= ($glade->get_widget('additional_checkbutton')->get_active	? 1 : 0);
	$OPTIONS->{authoritative_on}	= ($glade->get_widget('authoritative_checkbutton')->get_active	? 1 : 0);
	$OPTIONS->{trace_on}		= ($glade->get_widget('trace_checkbutton')->get_active		? 1 : 0);
	$OPTIONS->{options_visible}	= ($glade->get_widget('options_expander')->get_expanded		? 1 : 0);
	$OPTIONS->{current_page}	= $glade->get_widget('notebook')->get_current_page;
	$OPTIONS->{last_type}		= $glade->get_widget('type_combo')->get_active;
	$OPTIONS->{query_hist}		= join('|', uniq(@QUERY_HIST));
	$OPTIONS->{server_hist}		= join('|', uniq(@SERVER_HIST));
	save_options($OPTIONS);
	exit 0;
}

### make a DNS query:
sub make_query {

	### query options:
	my $type		= $glade->get_widget('type_combo')->get_model->get($glade->get_widget('type_combo')->get_active_iter, 0);
	my $query		= $glade->get_widget('query_combo')->entry->get_text;
	my $server		= $glade->get_widget('server_combo')->entry->get_text;
	my $recursive		= $glade->get_widget('recursive_checkbutton')->get_active;
	my $additional		= $glade->get_widget('additional_checkbutton')->get_active;
	my $authoritative	= $glade->get_widget('authoritative_checkbutton')->get_active;
	my $trace		= $glade->get_widget('trace_checkbutton')->get_active;

	return unless ($query ne '');

	unshift(@QUERY_HIST, $query);
	@QUERY_HIST = uniq(@QUERY_HIST);
	unshift(@SERVER_HIST, $server) if ($server ne '');
	@SERVER_HIST = uniq(@SERVER_HIST);

	$glade->get_widget('query_combo')->set_popdown_strings(@QUERY_HIST);
	$glade->get_widget('server_combo')->set_popdown_strings(@SERVER_HIST) if ($server ne '');
	$glade->get_widget('query_button')->set_sensitive(0);

	### if we're doing a PTR query and the query string is a dotted-quad IP address,
	### then turn it into an in-addr.arpa domain:
	if ($type eq 'PTR') {
		if ($query =~ /^\d+\.\d+\.\d+\.\d+$/) {
			$query = join('.', reverse(split(/\./, $query))).'.in-addr.arpa';
			$glade->get_widget('query_combo')->entry->set_text($query);
		}
	}

	### build the dig query:
	my $cmd = join(' ', (
		$dig,
		'+'.($recursive		== 1 ? '' : 'no').'recursive',
		'+'.($additional	== 1 ? '' : 'no').'additional',
		'+'.($authoritative	== 1 ? '' : 'no').'authority',
		'+'.($trace		== 1 ? '' : 'no').'trace',
		$type,
		$query,
		($server ne '' ? sprintf('@%s', $server) : ''),
	));

	$glade->get_widget('main_window')->window->set_cursor($busy);
	$store->clear;
	$glade->get_widget('raw_response')->get_buffer->set_text('');

	### read dig's STDOUT, pipe STDERR to a file so we can display it
	### if dig exits with an error:
	my $tmpfile = tmpnam();
	if (!open(DIG, "$cmd 2>$tmpfile|")) {
		$glade->get_widget('main_window')->window->set_cursor($normal);
		$glade->get_widget('query_button')->set_sensitive(1);

		my $dialog = Gtk2::MessageDialog->new(undef, 'modal', 'error', 'ok', sprintf(gettex('Cannot pipe from dig: %s'), $!));
		$dialog->signal_connect('response', sub { $dialog->destroy });
		$dialog->run;

	} else {
		my ($tag, $buffer);
		$tag = Gtk2::Helper->add_watch(fileno(DIG), 'in', sub {
			if (eof(DIG)) {
				close(DIG);
				Gtk2::Helper->remove_watch($tag);
				display($buffer);
				$glade->get_widget('main_window')->window->set_cursor($normal);
				$glade->get_widget('query_button')->set_sensitive(1);
				if ($? > 0) {
					if (!open(TMPFILE, $tmpfile)) {
						print STDERR "Error: couldn't open temporary file '$tmpfile' to tell you about an error from dig: $!\n";
						exit 1;

					} else {
						my $err;
						while (<TMPFILE>) {
							$err .= $_;
						}
						close(TMPFILE);

						my $dialog = Gtk2::MessageDialog->new(undef, 'modal', 'error', 'ok', $err);
						$dialog->signal_connect('response', sub { $dialog->destroy });
						$dialog->run;
					}

				} else {
					unlink($tmpfile);

				}
			} else {
				$buffer .= <DIG>;

			}
		});
	}

	return 1;
}

sub display {
	my $buffer = shift;

	my @lines = split(/\n{1}/, $buffer);
	my $parent = undef;

	$glade->get_widget('raw_response')->get_buffer->insert_with_tags_by_name(
		$glade->get_widget('raw_response')->get_buffer->get_start_iter,
		$buffer,
		'monospace',
	);

	foreach my $line (@lines) {
		if ($line =~ /;; ANSWER SECTION/) {
			my $iter = $store->append(undef);
			$store->set($iter, 0, 'Answer:');
			$parent = $iter;

		} elsif ($line =~ /;; ADDITIONAL SECTION/) {
			my $iter = $store->append(undef);
			$store->set($iter, 0, 'Additional:');
			$parent = $iter;

		} elsif ($line =~ /;; AUTHORITY SECTION/) {
			my $iter = $store->append(undef);
			$store->set($iter, 0, 'Authority:');
			$parent = $iter;

		} elsif ($line !~ /^\s*$/ && $line !~ /^\s*;/) {
			my $iter = $store->append($parent);
			my @parts = split(/\s+/, $line, 5);

			for (0..4) {
				$store->set($iter, $_, $parts[$_]);
			}

		}
	}
	$glade->get_widget('output')->expand_all;
	return 1;
}

sub load_options {
	my $hashref = {
		font			=> 'Courier 10',
		recursive_on		=> 1,
		additional_on		=> 1,
		authoritative_on	=> 1,
		trace_on		=> 0,
		options_visible		=> 0,
		current_page		=> 0,
		last_type		=> 0,
	};
	if (open(RCFILE, $RCFILE)) {
		while (<RCFILE>) {
			chomp;
			my ($name, $value) = split(/\s*=\s*/, $_, 2);
			$hashref->{lc($name)} = $value;
		}
		close(RCFILE);
	}
	return $hashref;
}

sub save_options {
	my $hashref = shift;
	if (!open(RCFILE, ">$RCFILE")) {
		print STDERR "Error opening '$RCFILE': $!\n";
		return undef;

	} else {
		foreach my $name (keys(%{$hashref})) {
			printf(RCFILE "%s=%s\n", $name, $hashref->{$name});
		}
		close(RCFILE);
		return 1;
	}
}

sub uniq {
	my @array = @_;
	my @new;
	my %map;
	foreach my $member (@array) {
		$map{$member}++;
	}
	foreach my $member (@array) {
		if ($map{$member} > 0) {
			push(@new, $member);
			$map{$member} = 0;
		}
	}
	return @new;
}

sub show_about_dialog {
	$glade->get_widget('about_dialog')->set_position('center');
	$glade->get_widget('about_dialog')->show_all;
	return 1;
}

sub hide_about_dialog {
	$glade->get_widget('about_dialog')->hide_all;
	return 1;
}

sub changed {
	$glade->get_widget('query_button')->set_sensitive($glade->get_widget('query_combo')->entry->get_text ne '');
	return 1;
}
